import uuid

from datetime import date, datetime
from decimal import Decimal

import brewman.schemas.input as schema_in
import brewman.schemas.voucher as output

from fastapi import APIRouter, Depends, File, HTTPException, Request, Security, status
from sqlalchemy import distinct, func, or_, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

from ..core.security import get_current_active_user as get_user
from ..core.session import get_date, set_date
from ..db.session import SessionFuture
from ..models.account_base import AccountBase
from ..models.batch import Batch
from ..models.inventory import Inventory
from ..models.journal import Journal
from ..models.product import Product
from ..models.rate_contract import RateContract
from ..models.rate_contract_item import RateContractItem
from ..models.stock_keeping_unit import StockKeepingUnit
from ..models.validations import check_duplicate_skus, check_journals_are_valid
from ..models.voucher import Voucher
from ..models.voucher_type import VoucherType
from ..schemas.blank_voucher_info import BlankVoucherInfo
from ..schemas.inventory import Inventory as InventorySchema
from ..schemas.user import UserToken
from . import get_lock_info
from .batch_integrity import fix_single_batch_prices
from .db_image import save_files, update_files
from .voucher import (
    blank_voucher,
    check_voucher_edit_allowed,
    get_batch_quantity,
    voucher_info,
)


router = APIRouter()


@router.post("", response_model=output.Voucher)
def save_route(
    request: Request,
    data: schema_in.PurchaseIn = Depends(schema_in.PurchaseIn.load_form),
    i: list[bytes] = File(None),
    t: list[bytes] = File(None),
    user: UserToken = Security(get_user, scopes=["purchase"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = save(data, user, db)
            save_inventories(item, data.vendor.id_, data.inventories, db)
            check_duplicate_skus(item)
            save_journals(item, data.vendor, db)
            check_journals_are_valid(item)
            save_files(item.id, i, t, db)
            db.commit()
            set_date(data.date_.strftime("%d-%b-%Y"), request.session)
            info = voucher_info(item, db)
            return info
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def save(data: schema_in.PurchaseIn, user: UserToken, db: Session) -> Voucher:
    product_accounts = (
        select(Product.account_id)
        .join(Product.skus)
        .where(StockKeepingUnit.id.in_([i.batch.sku.id_ for i in data.inventories]))
    )
    account_types: list[int] = list(
        db.execute(
            select(distinct(AccountBase.type_id)).where(
                or_(
                    AccountBase.id == data.vendor.id_,
                    AccountBase.id.in_(product_accounts),
                )
            )
        )
        .scalars()
        .all()
    )
    allowed, message = get_lock_info([data.date_], VoucherType[data.type_.replace(" ", "_").upper()], account_types, db)
    if not allowed:
        raise HTTPException(
            status_code=status.HTTP_423_LOCKED,
            detail=message,
        )
    voucher = Voucher(
        date_=data.date_,
        narration=data.narration,
        is_starred=data.is_starred,
        user_id=user.id_,
        voucher_type=VoucherType[data.type_.replace(" ", "_").upper()],
    )
    db.add(voucher)
    return voucher


def save_inventories(voucher: Voucher, vendor_id: uuid.UUID, inventories: list[InventorySchema], db: Session) -> None:
    for item in inventories:
        sku: StockKeepingUnit = db.execute(
            select(StockKeepingUnit).where(StockKeepingUnit.id == item.batch.sku.id_)
        ).scalar_one()
        rc_price = rate_contract_price(sku.id, vendor_id, voucher.date_, db)
        if rc_price is not None and rc_price != item.rate:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Product price does not match the Rate Contract price",
            )
        if rc_price is not None:
            item.tax = Decimal(0)
            item.discount = Decimal(0)
        batch = Batch(
            name=voucher.date_,
            sku=sku,
            quantity_remaining=item.quantity,
            rate=item.rate,
            tax=item.tax,
            discount=item.discount,
        )
        db.add(batch)
        inventory = Inventory(
            id_=item.id_,
            batch=batch,
            quantity=item.quantity,
            rate=item.rate,
            tax=item.tax,
            discount=item.discount,
        )
        sku.cost_price = item.rate
        voucher.inventories.append(inventory)
        db.add(inventory)


def save_journals(voucher: Voucher, ven: schema_in.AccountLink, db: Session) -> None:
    vendor = db.execute(select(AccountBase).where(AccountBase.id == ven.id_)).scalar_one()
    journals: dict[uuid.UUID, Journal] = {}
    amount = Decimal(0)
    for i_item in voucher.inventories:
        account_id, cc_id = db.execute(
            select(AccountBase.id, AccountBase.cost_centre_id)
            .join(AccountBase.products)
            .join(Product.skus)
            .where(StockKeepingUnit.id == i_item.batch.sku.id)
        ).one()
        amount += round(i_item.amount, 2)
        if account_id in journals:
            journals[account_id].amount += round(i_item.amount, 2)
        else:
            journals[account_id] = Journal(
                debit=1,
                cost_centre_id=cc_id,
                account_id=account_id,
                amount=round(i_item.amount, 2),
            )
    journals[vendor.id] = Journal(
        debit=-1,
        cost_centre_id=vendor.cost_centre_id,
        account_id=vendor.id,
        amount=amount,
    )
    for item in journals.values():
        voucher.journals.append(item)
        db.add(item)


@router.put("/{id_}", response_model=output.Voucher)
def update_route(
    id_: uuid.UUID,
    request: Request,
    data: schema_in.PurchaseIn = Depends(schema_in.PurchaseIn.load_form),
    i: list[bytes] = File(None),
    t: list[bytes] = File(None),
    user: UserToken = Security(get_user, scopes=["purchase"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = update_voucher(id_, data, user, db)
            update_inventory(item, data.vendor.id_, data.inventories, db)
            check_duplicate_skus(item)
            update_journals(item, data.vendor, db)
            check_journals_are_valid(item)
            update_files(item.id, data.files, i, t, db)
            db.commit()
            set_date(data.date_.strftime("%d-%b-%Y"), request.session)
            return voucher_info(item, db)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def update_voucher(id_: uuid.UUID, data: schema_in.PurchaseIn, user: UserToken, db: Session) -> Voucher:
    voucher: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
    product_accounts = (
        select(Product.account_id)
        .join(Product.skus)
        .where(StockKeepingUnit.id.in_([i.batch.sku.id_ for i in data.inventories]))
    )
    account_types: list[int] = list(
        db.execute(
            select(distinct(AccountBase.type_id)).where(
                or_(
                    AccountBase.id.in_([data.vendor.id_] + [vj.account_id for vj in voucher.journals]),
                    AccountBase.id.in_(product_accounts),
                )
            )
        )
        .scalars()
        .all()
    )
    allowed, message = get_lock_info([voucher.date_, data.date_], voucher.voucher_type, account_types, db)
    if not allowed:
        raise HTTPException(
            status_code=status.HTTP_423_LOCKED,
            detail=message,
        )

    check_voucher_edit_allowed(voucher, user)
    voucher.date_ = data.date_
    voucher.is_starred = data.is_starred
    voucher.narration = data.narration
    voucher.user_id = user.id_
    voucher.posted = False
    voucher.last_edit_date = datetime.utcnow()
    return voucher


def update_inventory(voucher: Voucher, vendor_id: uuid.UUID, inventories: list[InventorySchema], db: Session) -> None:
    for it in range(len(voucher.inventories), 0, -1):
        item = voucher.inventories[it - 1]
        quantity_consumed = -1 * get_batch_quantity(item.batch_id, voucher.id, db)
        index = next(
            (idx for (idx, d) in enumerate(inventories) if d.id_ == item.id and d.batch.sku.id_ == item.batch.sku_id),
            None,
        )
        if index is not None:
            new_inventory = inventories.pop(index)
            sku: StockKeepingUnit = db.execute(
                select(StockKeepingUnit).where(StockKeepingUnit.id == item.batch.sku.id)
            ).scalar_one()
            rc_price = rate_contract_price(sku.id, vendor_id, voucher.date_, db)
            if batch_has_older_vouchers(item.batch_id, voucher.date_, voucher.id, db):
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=f"{item.batch.sku.product.name} has older vouchers",
                )
            if rc_price is not None and rc_price != new_inventory.rate:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Product price does not match the Rate Contract price",
                )
            if rc_price is not None:
                new_inventory.tax = Decimal(0)
                new_inventory.discount = Decimal(0)
            if new_inventory.quantity < quantity_consumed:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=f"{quantity_consumed} is the minimum as it has been issued",
                )
            item.batch.quantity_remaining = new_inventory.quantity - quantity_consumed
            item.quantity = new_inventory.quantity
            item.batch.name = voucher.date_
            item.rate = new_inventory.rate
            item.batch.rate = new_inventory.rate
            item.discount = new_inventory.discount
            item.batch.discount = new_inventory.discount
            item.tax = new_inventory.tax
            item.batch.tax = new_inventory.tax
            sku.cost_price = new_inventory.rate
            db.flush()
            fix_single_batch_prices(item.batch_id, db)
        else:
            has_been_issued: int = db.execute(
                select(func.count(Inventory.id)).where(Inventory.batch_id == item.batch.id, Inventory.id != item.id)
            ).scalar_one()
            if has_been_issued > 0:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=f"{item.batch.sku.product.name} has been issued, it cannot be deleted",
                )
            else:
                db.delete(item.batch)
                db.delete(item)
                voucher.inventories.remove(item)
    save_inventories(voucher, vendor_id, inventories, db)


def update_journals(voucher: Voucher, ven: schema_in.AccountLink, db: Session) -> None:
    vendor = db.execute(select(AccountBase).where(AccountBase.id == ven.id_)).scalar_one()
    journals: dict[uuid.UUID, Journal] = {}
    amount = Decimal(0)
    for i_item in voucher.inventories:
        account_id, cc_id = db.execute(
            select(AccountBase.id, AccountBase.cost_centre_id)
            .join(AccountBase.products)
            .join(Product.skus)
            .where(StockKeepingUnit.id == i_item.batch.sku.id)
        ).one()
        amount += i_item.amount
        if account_id in journals:
            journals[account_id].amount += i_item.amount
        else:
            journals[account_id] = Journal(
                debit=1,
                cost_centre_id=cc_id,
                account_id=account_id,
                amount=i_item.amount,
            )
    journals[vendor.id] = Journal(
        debit=-1,
        cost_centre_id=vendor.cost_centre_id,
        account_id=vendor.id,
        amount=amount,
    )
    for i in range(len(voucher.journals), 0, -1):
        item = voucher.journals[i - 1]
        if item.account_id in journals:
            item.debit = journals[item.account_id].debit
            item.amount = round(journals[item.account_id].amount, 2)
            item.cost_centre_id = journals[item.account_id].cost_centre_id
            del journals[item.account_id]
        else:
            db.delete(item)
            voucher.journals.remove(item)
    for item in journals.values():
        item.amount = round(item.amount, 2)
        voucher.journals.append(item)
        db.add(item)


@router.get("/{id_}", response_model=output.Voucher)
def get_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["purchase"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
            return voucher_info(item, db)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.get("", response_model=output.Voucher)
def show_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["purchase"]),
) -> output.Voucher:
    additional_info = BlankVoucherInfo(date_=get_date(request.session), type_=VoucherType.PURCHASE)  # type: ignore
    with SessionFuture() as db:
        return blank_voucher(additional_info, db)


def rate_contract_price(id_: uuid.UUID, vendor_id: uuid.UUID, date_: date, db: Session) -> Decimal | None:
    contracts = select(RateContract.id).where(
        RateContract.vendor_id == vendor_id, RateContract.valid_from <= date_, RateContract.valid_till >= date_
    )
    return db.execute(
        select(RateContractItem.price).where(
            RateContractItem.id == id_, RateContractItem.rate_contract_id.in_(contracts)
        )
    ).scalar_one_or_none()


def batch_has_older_vouchers(id_: uuid.UUID, date_: date, voucher_id: uuid.UUID, db: Session) -> bool:
    count_: int = db.execute(
        select(func.count())
        .join(Batch.inventories)
        .join(Inventory.voucher)
        .where(Batch.id == id_, Voucher.date_ < date_, Voucher.id != voucher_id)
    ).scalar_one()
    return count_ > 0
