import datetime
import uuid

import brewman.schemas.recipe_template as schemas

from fastapi import APIRouter, HTTPException, Security, status
from sqlalchemy import delete, select, update
from sqlalchemy.exc import SQLAlchemyError

from ..core.security import get_current_active_user as get_user
from ..db.session import SessionFuture
from ..models.recipe_template import RecipeTemplate
from ..schemas.user import UserToken


router = APIRouter()


@router.post("", response_model=None)
def save(
    data: schemas.RecipeTemplateIn,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> None:
    try:
        with SessionFuture() as db:
            date_ = data.date_ if data.date_ is not None else datetime.date.today()
            if data.selected:
                db.execute(update(RecipeTemplate).values(selected=False))
            item = RecipeTemplate(id=None, name=data.name, date_=date_, text=data.text, selected=data.selected)
            db.add(item)
            db.commit()
            return None
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.put("/{id_}", response_model=None)
def update_route(
    id_: uuid.UUID,
    data: schemas.RecipeTemplateIn,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> None:
    try:
        with SessionFuture() as db:
            date_ = data.date_ if data.date_ is not None else datetime.date.today()
            item = db.execute(select(RecipeTemplate).where(RecipeTemplate.id == id_)).scalar_one()
            if data.selected and not item.selected:
                db.execute(update(RecipeTemplate).where(RecipeTemplate.id != id_).values(selected=False))
            item.name = data.name
            item.text = data.text
            item.selected = data.selected
            item.date_ = date_
            db.commit()
            return None
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.delete("/{id_}", response_model=None)
def delete_route(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> None:
    try:
        with SessionFuture() as db:
            db.execute(delete(RecipeTemplate).where(RecipeTemplate.id == id_))
            db.commit()
            return None
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.get("", response_model=schemas.RecipeTemplateIn)
def show_blank(
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.RecipeTemplateIn:
    return schemas.RecipeTemplateIn(name="", date_=datetime.date.today(), text="", selected=False)


@router.get("/list", response_model=list[schemas.RecipeTemplate])
async def show_list(
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> list[schemas.RecipeTemplate]:
    with SessionFuture() as db:
        list_ = db.execute(select(RecipeTemplate).order_by(RecipeTemplate.name)).scalars().all()
        return [
            schemas.RecipeTemplate(id_=i.id, name=i.name, date_=i.date_, text=i.text, selected=i.selected)
            for i in list_
        ]


@router.get("/{id_}", response_model=schemas.RecipeTemplate)
def show_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.RecipeTemplate:
    with SessionFuture() as db:
        item = db.execute(select(RecipeTemplate).where(RecipeTemplate.id == id_)).scalar_one()
        return schemas.RecipeTemplate(
            id_=item.id, name=item.name, date_=item.date_, text=item.text, selected=item.selected
        )
