import uuid

from datetime import date
from decimal import Decimal
from typing import Sequence

import brewman.schemas.voucher as output

from fastapi import APIRouter, HTTPException, Security, status
from sqlalchemy import and_, distinct, func, or_, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

from ..core.security import get_current_active_user as get_user
from ..core.session import get_first_day
from ..db.session import SessionFuture
from ..models.account import Account
from ..models.account_base import AccountBase
from ..models.attendance import Attendance
from ..models.attendance_type import AttendanceType
from ..models.cost_centre import CostCentre
from ..models.db_image import DbImage
from ..models.employee import Employee
from ..models.inventory import Inventory
from ..models.journal import Journal
from ..models.voucher import Voucher
from ..models.voucher_type import VoucherType
from ..routers import get_lock_info
from ..schemas.blank_voucher_info import BlankVoucherInfo
from ..schemas.user import UserToken


router = APIRouter()


@router.post("/post-voucher/{id_}", response_model=output.Voucher)
def post_voucher(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["post-vouchers"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            voucher: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
            account_types: Sequence[int] = (
                db.execute(
                    select(distinct(AccountBase.type_id)).where(
                        AccountBase.id.in_([vj.account_id for vj in voucher.journals])
                    )
                )
                .scalars()
                .all()
            )
            allowed, message = get_lock_info([voucher.date_], voucher.voucher_type, account_types, db)
            if not allowed:
                raise HTTPException(
                    status_code=status.HTTP_423_LOCKED,
                    detail=message,
                )
            voucher.posted = True
            voucher.poster_id = user.id_
            db.commit()
            return voucher_info(voucher, db)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def check_delete_permissions(voucher: Voucher, user: UserToken) -> None:
    voucher_type = voucher.voucher_type.name.replace("_", "-").lower()
    if voucher_type in ["payment", "receipt"]:
        voucher_type = "journal"
    if voucher.posted and "edit-posted-vouchers" not in user.permissions:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="You are not allowed to edit posted vouchers",
        )
    elif voucher.user_id != user.id_ and "edit-other-user's-vouchers" not in user.permissions:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="You are not allowed to edit other user's vouchers",
        )
    elif voucher_type not in user.permissions:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=f"You are not allowed ({voucher.voucher_type.name.replace('_', ' ').title()}) vouchers",
        )


@router.delete("/delete/{id_}")
def delete_voucher(
    id_: uuid.UUID,
    user: UserToken = Security(get_user),
) -> output.Voucher:
    with SessionFuture() as db:
        voucher: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
        images = db.execute(select(DbImage).where(DbImage.resource_id == voucher.id)).scalars().all()
        check_delete_permissions(voucher, user)
        account_types: Sequence[int] = (
            db.execute(
                select(distinct(AccountBase.type_id)).where(
                    AccountBase.id.in_([vj.account_id for vj in voucher.journals])
                )
            )
            .scalars()
            .all()
        )
        allowed, message = get_lock_info([voucher.date_], voucher.voucher_type, account_types, db)
        if not allowed:
            raise HTTPException(
                status_code=status.HTTP_423_LOCKED,
                detail=message,
            )
        json_voucher = voucher_info(voucher, db)
        batches_to_delete = []
        if voucher.voucher_type == VoucherType.ISSUE:
            for j_item in voucher.journals:
                if j_item.debit == 1:
                    destination = j_item.cost_centre_id
                else:
                    source = j_item.cost_centre_id

            batch_consumed: bool | None
            if source == CostCentre.cost_centre_purchase():
                batch_consumed = True
            elif destination == CostCentre.cost_centre_purchase():
                batch_consumed = False
            else:
                batch_consumed = None

            if batch_consumed is None:
                pass
            elif batch_consumed:
                for i_item in voucher.inventories:
                    i_item.batch.quantity_remaining += i_item.quantity
            else:
                for i_item in voucher.inventories:
                    if i_item.batch.quantity_remaining < i_item.quantity:
                        raise HTTPException(
                            status_code=status.HTTP_423_LOCKED,
                            detail=f"Only {i_item.batch.quantity_remaining}"
                            f" of {i_item.batch.sku.product.name} remaining.\n"
                            f"So it cannot be deleted",
                        )
                    i_item.batch.quantity_remaining -= i_item.quantity
        elif voucher.voucher_type == VoucherType.PURCHASE:
            for i_item in voucher.inventories:
                uses: int = db.execute(
                    select(func.count(Inventory.id)).where(
                        Inventory.batch_id == i_item.batch.id, Inventory.id != i_item.id
                    )
                ).scalar_one()
                if uses > 0:
                    raise HTTPException(
                        status_code=status.HTTP_423_LOCKED,
                        detail=f"{i_item.batch.sku.product.name} has been issued and cannot be deleted",
                    )
                batches_to_delete.append(i_item.batch)
        elif voucher.voucher_type == VoucherType.PURCHASE_RETURN:
            for item in voucher.inventories:
                item.batch.quantity_remaining += item.quantity
        for b in batches_to_delete:
            db.delete(b)
        db.delete(voucher)
        for image in images:
            db.delete(image)
        db.commit()

        return blank_voucher(
            info=BlankVoucherInfo(
                date_=json_voucher.date_,
                type_=VoucherType[json_voucher.type_],
                source=json_voucher.source,
                destination=json_voucher.destination,
            ),
            db=db,
        )


def voucher_info(voucher: Voucher, db: Session) -> output.Voucher:
    json_voucher = output.Voucher(
        id_=voucher.id,
        date_=voucher.date_,
        is_starred=voucher.is_starred,
        type_=voucher.voucher_type.name,
        posted=voucher.posted,
        narration=voucher.narration,
        journals=[],
        inventories=[],
        employee_benefits=[],
        incentives=[],
        incentive=Decimal(0),
        files=[],
        creation_date=voucher.creation_date,
        last_edit_date=voucher.last_edit_date,
        user=output.UserLink(id_=voucher.user.id, name=voucher.user.name),
        poster=voucher.poster.name if voucher.posted else "",
    )
    if voucher.reconcile_date is not None:
        json_voucher.reconcile_date = voucher.reconcile_date
    if voucher.voucher_type == VoucherType.PURCHASE:
        item = [j for j in voucher.journals if j.debit == -1][0]
        json_voucher.vendor = output.AccountLink(id_=item.account.id, name=item.account.name)
    elif voucher.voucher_type == VoucherType.ISSUE:
        item = [j for j in voucher.journals if j.debit == -1][0]
        json_voucher.source = output.CostCentreLink(id_=item.cost_centre_id, name="")
        item = [j for j in voucher.journals if j.debit == 1][0]
        json_voucher.destination = output.CostCentreLink(id_=item.cost_centre_id, name="")
    if voucher.voucher_type == VoucherType.PURCHASE_RETURN:
        item = [j for j in voucher.journals if j.debit == 1][0]
        json_voucher.vendor = output.AccountLink(id_=item.account.id, name=item.account.name)
    else:
        for journal in voucher.journals:
            json_voucher.journals.append(
                output.Journal(
                    id_=journal.id,
                    debit=journal.debit,
                    amount=journal.amount,
                    account=output.AccountLink(id_=journal.account.id, name=journal.account.name),
                    cost_centre=output.CostCentreLink(id_=journal.cost_centre_id),
                )
            )
    for benefit in voucher.employee_benefits:
        json_voucher.employee_benefits.append(
            output.EmployeeBenefit(
                gross_salary=benefit.gross_salary,
                days_worked=benefit.days_worked,
                esi_employee=benefit.esi_ee,
                pf_employee=benefit.pf_ee,
                esi_employer=benefit.esi_er,
                pf_employer=benefit.pf_er,
                employee=output.EmployeeLink(
                    id_=benefit.journal.account.id,
                    name=benefit.journal.account.name,
                    designation=benefit.journal.account.designation,
                    cost_centre=output.CostCentreLink(
                        id_=benefit.journal.account.cost_centre.id,
                        name=benefit.journal.account.cost_centre.name,
                    ),
                ),
            )
        )
    for incentive in voucher.incentives:
        employee: Employee = db.execute(
            select(Employee).where(Employee.id == incentive.journal.account_id)
        ).scalar_one()
        json_voucher.incentives.append(
            output.Incentive(
                employee_id=incentive.journal.account_id,
                name=incentive.journal.account.name,
                designation=employee.designation,
                department=incentive.journal.account.cost_centre.name,
                days_worked=incentive.days_worked,
                points=incentive.points,
            )
        )
    if len(json_voucher.incentives) > 0:
        json_voucher.incentive = next(x.amount for x in voucher.journals if x.account_id == Account.incentive_id())
    for inventory in voucher.inventories:
        text = (
            f"{inventory.batch.sku.product.name} ({inventory.batch.sku.units}) "
            f"{inventory.batch.quantity_remaining:.2f}@{inventory.batch.rate:.2f} "
            f"from {inventory.batch.name.strftime('%d-%b-%Y')}"
        )
        json_voucher.inventories.append(
            output.Inventory(
                id_=inventory.id,
                quantity=inventory.quantity,
                rate=inventory.rate,
                tax=inventory.tax,
                discount=inventory.discount,
                amount=inventory.amount,
                batch=output.Batch(
                    id_=inventory.batch.id,
                    name=text,
                    quantity_remaining=inventory.batch.quantity_remaining,
                    tax=inventory.batch.tax,
                    discount=inventory.batch.discount,
                    rate=inventory.batch.rate,
                    sku=output.ProductLink(
                        id_=inventory.batch.sku.id,
                        name=f"{inventory.batch.sku.product.name} ({inventory.batch.sku.units})"
                        if inventory.batch.sku.units
                        else inventory.batch.sku.product.name,
                    ),
                ),
            )
        )
    images = db.execute(select(DbImage).where(DbImage.resource_id == voucher.id)).scalars().all()
    for image in images:
        json_voucher.files.append(
            output.ImageUpload(
                id_=image.id,
                resized=f"/db-image/{image.id}/resized",
                thumbnail=f"/db-image/{image.id}/thumbnail",
            )
        )
    return json_voucher


def blank_voucher(info: BlankVoucherInfo, db: Session) -> output.Voucher:
    json_voucher = output.Voucher(
        type_=info.type_.name,
        date_=info.date_,
        is_starred=False,
        posted=False,
        narration="",
        journals=[],
        inventories=[],
        incentives=[],
        employee_benefits=[],
        files=[],
    )
    if info.type_ == VoucherType.JOURNAL:
        pass
    elif info.type_ == VoucherType.PAYMENT:
        account = None
        if info.account is not None:
            account = db.execute(select(AccountBase).where(AccountBase.id == info.account.id_)).scalars().one_or_none()
        if account is not None:
            j_account = output.AccountLink(id_=account.id, name=account.name)
        else:
            j_account = AccountBase.cash_in_hand()
        json_voucher.journals.append(output.Journal(account=j_account, amount=Decimal(0), debit=-1))
    elif info.type_ == VoucherType.RECEIPT:
        account = None
        if info.account is not None:
            account = db.execute(select(AccountBase).where(AccountBase.id == info.account.id_)).scalars().one_or_none()
        if account is not None:
            j_account = output.AccountLink(id_=account.id, name=account.name)
        else:
            j_account = AccountBase.cash_in_hand()
        json_voucher.journals.append(output.Journal(account=j_account, amount=Decimal(0), debit=1))
    elif info.type_ == VoucherType.PURCHASE:
        json_voucher.vendor = AccountBase.local_purchase()
    elif info.type_ == VoucherType.PURCHASE_RETURN:
        json_voucher.vendor = AccountBase.local_purchase()
    elif info.type_ == VoucherType.ISSUE:
        if info.source is not None:
            json_voucher.source = output.CostCentreLink(id_=info.source.id_)
        else:
            json_voucher.source = output.CostCentreLink(id_=CostCentre.cost_centre_purchase())
        if info.destination is not None:
            json_voucher.destination = output.CostCentreLink(id_=info.destination.id_)
        else:
            json_voucher.destination = output.CostCentreLink(id_=CostCentre.cost_centre_kitchen())
    elif info.type_ == VoucherType.EMPLOYEE_BENEFIT:
        pass
    elif info.type_ == VoucherType.INCENTIVE:
        json_voucher.incentives, json_voucher.incentive = incentive_employees(info.date_, db)
    else:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f'Voucher of type "{info.type_.name}" does not exist.',
        )
    return json_voucher


def incentive_employees(date_: date, db: Session) -> tuple[list[output.Incentive], Decimal]:
    start_date = get_first_day(date_)
    finish_date = date_
    details: list[output.Incentive] = []
    employees = (
        db.execute(
            select(Employee)
            .where(
                Employee.joining_date <= finish_date,
                or_(Employee.is_active == True, Employee.leaving_date >= start_date),  # noqa: E712
            )
            .order_by(Employee.cost_centre_id, Employee.designation, Employee.name)
        )
        .scalars()
        .all()
    )
    for employee in employees:
        att: Sequence[Attendance] = (
            db.execute(
                select(Attendance).where(
                    Attendance.employee_id == employee.id,
                    Attendance.date_ >= start_date,
                    Attendance.date_ <= finish_date,
                    Attendance.is_valid == True,  # noqa: E712
                )
            )
            .scalars()
            .all()
        )
        days_worked = sum(AttendanceType.by_id(x.attendance_type).value for x in att) or Decimal(0)
        details.append(
            output.Incentive(
                employee_id=employee.id,
                name=employee.name,
                designation=employee.designation,
                department=employee.cost_centre.name,
                days_worked=days_worked,
                points=employee.points,
            )
        )

    amount: Decimal | None = db.execute(
        select(func.sum(Journal.amount * Journal.debit))
        .join(Journal.voucher)
        .where(
            Journal.account_id == Account.incentive_id(),
            Voucher.voucher_type != VoucherType.ISSUE,
            or_(
                Voucher.date_ <= finish_date,
                and_(
                    Voucher.date_ == finish_date,
                    Voucher.voucher_type != VoucherType.INCENTIVE,
                ),
            ),
        )
    ).scalar()
    amount = Decimal(0) if amount is None else amount * Decimal(0.9) * -1
    return details, amount


def check_voucher_edit_allowed(voucher: Voucher, user: UserToken) -> None:
    if voucher.posted and "edit-posted-vouchers" not in user.permissions:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="You are not allowed to edit posted vouchers",
        )
    elif voucher.user_id != user.id_ and "edit-other-user's-vouchers" not in user.permissions:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="You are not allowed to edit other user's vouchers",
        )


def get_batch_quantity(id_: uuid.UUID, voucher_id: uuid.UUID | None, db: Session) -> Decimal:
    query = (
        select(func.sum(Inventory.quantity * Journal.debit))
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .where(Inventory.batch_id == id_, Journal.cost_centre_id == CostCentre.cost_centre_purchase())
    )
    if voucher_id is not None:
        query = query.where(Voucher.id != voucher_id)
    return db.execute(query).scalar_one() or Decimal(0)
